#ifndef _UDM_UTILS_H
#define _UDM_UTILS_H

/* for time_t */
#include <time.h>

/* for va_list */
#include <stdarg.h>

#include "udm_os.h"

/* This is used in UdmTime_t2Str and in its callers */
#define UDM_MAXTIMESTRLEN	35


/* Some useful MACROs */
#define UDM_STRNCMP(x,y)	(strncmp(x,y,strlen(y)))
#define UDM_STRNCASECMP(x,y)	(strncasecmp(x,y,strlen(y)))
#define UDM_STREND(s)		(s+strlen(s))
#define UDM_FREE(x)		{if(x){free(x);x=NULL;}}
#define UDM_STRNCPY(x,y)	{strncpy(x,y,sizeof(x));x[sizeof(x)-1]=0;}
#define UDM_SKIP(s,set)		while((*s)&&(strchr(set,*s)))s++;
#define UDM_SKIPN(s,set)	while((*s)&&(!strchr(set,*s)))s++;

/* NULL safe atoi*/
#define UDM_ATOI(x)		((x)?atoi(x):0)
#define UDM_NULL2EMPTY(x)	((x)?(x):"")
#define UDM_EMPTY2NULL(x)	((x)?((*(x))?(x):NULL):NULL)


/* Misc functions */
extern __INDLIB__ int		UdmInit();
extern char * UdmGetToken(char *s, const char *delim, char **last);
extern char * UdmTrim(char * p, char * delim);
extern char * UdmRTrim(char* p, char * delim);
extern char * UdmHtmlSpecialChars(char *str);
extern char * UdmUnescapeCGIQuery(char *d,char *s);
extern char * UdmEscapeURL(char *d,char *s);
extern char * UdmRemove2Dot(char *path);

/* This should convert Last-Modified time returned by webserver
 * to time_t (seconds since the Epoch). -kir
 */
extern time_t UdmHttpDate2Time_t(const char * date);

extern time_t UdmFTPDate2Time_t(char *date);

/***********************************************************
 * converts time_str to time_t (seconds)
 * time_str can be exactly number of seconds
 * or in the form 'xxxA[yyyB[zzzC]]'
 * (Spaces are allowed between xxx and A and yyy and so on)
 *   there xxx, yyy, zzz are numbers (can be negative!)
 *         A, B, C can be one of the following:
 *		s - second
 *		M - minute	
 *		h - hour
 *		d - day
 *		m - month
 *		y - year
 *	(these letters are as in strptime/strftime functions)
 *
 * Examples:
 * 1234 - 1234 seconds
 * 4h30M - 4 hours and 30 minutes (will return 9000 seconds)
 * 1y6m-15d - 1 year and six month minus 15 days (will return 45792000 s)
 * 1h-60M+1s - 1 hour minus 60 minutes plus 1 second (will return 1 s)
 */
time_t Udm_dp2time_t(char * time_str);


/* This one for printing HTTP Last-Modified: header */
extern void UdmTime_t2HttpStr(time_t t, char * time_str);
/* This one deals with timezone offset */
extern int UdmInitTZ(void);

extern unsigned long UdmStartTimer(void);

/* String functions */
int udm_strnlen(const char *s, int max);
/* returns length of string needed for vsnprintf */
int udm_vsnprintf_len(const char *fmt, va_list args);
int udm_vsnprintf(char *str, size_t size, const char  *fmt,  va_list ap);
int udm_snprintf(char *str, int size, const char *fmt, ...);


#define BASE64_LEN(len) (4 * (((len) + 2) / 3) +2)
extern void udm_base64_encode (const char *s, char *store, int length);
extern char * udm_rfc1522_decode(char * dst, char *src);

/* Build directory */
extern int   UdmBuild(char * path, int mode);


#endif /* _UDM_UTILS_H */
