#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>


#include "udm_config.h"
#include "udm_common.h"
#include "udm_charset.h"
#include "udm_filter.h"
#include "udm_utils.h"
#include "udm_parseurl.h"
#include "udm_log.h"
#include "udm_xmalloc.h"
#include "udm_robots.h"
#include "udm_server.h"
#include "udm_hrefs.h"
#include "udm_word.h"
#include "udm_parsehtml.h"


/**************************** Built-in Parsers ***************************/

int UdmParseText(UDM_AGENT * Indexer,UDM_SERVER * Server,
       char *content,int weight,int check_stopwords){
	char *s, *lt;
	if(weight&&content){
		s=UdmGetWord(content, &lt, Indexer->Conf->local_charset);
		while(s){
			UdmAddWord(Indexer, Server, s, weight, check_stopwords);
			s=UdmGetWord(NULL, &lt, Indexer->Conf->local_charset);
		}
	}
	return(0);
}


/* HTML parser states */
#define HTML_TAG	1
#define HTML_TXT	2
#define HTML_COM	3

int UdmParseHtml(UDM_AGENT *Indexer,UDM_SERVER *CurSrv,UDM_URL *CurURL,UDM_DOCUMENT *Doc){

	int inbody=0;
	int inscript=0;
	int instyle=0;
	int intitle=0;
	int hrefonly=0;
	int comment=0;
	int index=0;
	int follow=0;
	unsigned int group=0;
	char *htok;
	char str[UDMSTRSIZ]="";
	char res[UDMSTRSIZ]="";
	
	index=CurSrv->index;
	follow=CurSrv->follow;
	
	Doc->text=(char*)UdmXmalloc(2*UDM_MAXTEXTSIZE);
	Doc->keywords=(char*)UdmXmalloc(UDM_MAXKEYWORDSIZE);
	Doc->description=(char*)UdmXmalloc(UDM_MAXDESCSIZE);
	Doc->title=(char*)UdmXmalloc(UDM_MAXTITLESIZE);

	strcpy(str, CurURL->path);
	strcat(str, CurURL->filename);
	hrefonly=(UDM_HREFONLY==UdmFindFilter(Indexer->Conf,str,res));

	htok=Doc->content;

	while(*htok){
		char * href=NULL;
		char * tmp=NULL;
		char * hend;
		char * etmp;
		char * stmp;
		char * s;
		int have_space;
		int opening;
		UDM_TAG tag;
		int state=HTML_TXT;
		int len;


		if(!UDM_STRNCMP(htok,"<!--"))	state=HTML_COM;
		else	if(*htok=='<')		state=HTML_TAG;

		switch(state){
		case HTML_TAG: /* tag */

			href=NULL;
			for(hend=htok;(*hend!='>')&&(*hend);hend++);
			if(*hend=='>')hend++;
			tmp=(char*)malloc(hend-htok+1);
			strncpy(tmp,htok,hend-htok);tmp[hend-htok]=0;
			
			UdmParseTag(&tag,tmp);

			/* Convert to lower case */
			for(s=tag.tag;*s;*s=tolower(*s),s++);

			/* Detect whether opening or closing tag */
			if((opening=(tag.tag[0]!='/')))s=tag.tag;
			else				s=tag.tag+1;

			/* Let's find tag name in order of frequency */

			if(!strcmp(s,"a"))href=tag.href;/*117941*/
			else	if(!strcmp(s,"title"))	intitle=opening;/*6192*/
			else	
			if(!strcmp(s,"body")){/*5146*/
				if((inbody=opening)){
					int DCIndex;
#ifdef USE_CHARSET_GUESSER					
					DCIndex = UdmGuessCharset(Doc->content, UdmGetCharset(CurSrv->charset));
#else
					if(Indexer->charset){
						DCIndex=Indexer->charset;
					}else{
						DCIndex=UdmGetCharset(CurSrv->charset);
					}
#endif
					UdmRecode(Doc->content, DCIndex,Indexer->Conf->local_charset);
					UdmRecode(Doc->title,   DCIndex,Indexer->Conf->local_charset);
					UdmRecode(Doc->description,DCIndex,Indexer->Conf->local_charset);
					UdmRecode(Doc->keywords,DCIndex,Indexer->Conf->local_charset);
				}
			}else
			if((!strcmp(tag.tag,"meta"))&&(tag.name)&&(tag.content)){
			
				if((!strcasecmp(tag.name,"Content-Type"))&&(!Indexer->charset)){
					char *p;
					if((p=strstr(tag.content,"charset="))){
						Indexer->charset = UdmGetCharset(p + 8);
					}
				}else
				if(!strcasecmp(tag.name,"refresh")){
					if((href=strstr(tag.content,"URL=")))
						href+=4;
				}else
				if(!hrefonly&&!strcasecmp(tag.name,"keywords")){
					strncpy(Doc->keywords,tag.content,UDM_MAXKEYWORDSIZE-1);
					Doc->keywords[UDM_MAXKEYWORDSIZE-1]=0;
					
				}else
				if(!hrefonly&&!strcasecmp(tag.name,"description")){
					strncpy(Doc->description,tag.content,UDM_MAXDESCSIZE-1);
					Doc->description[UDM_MAXDESCSIZE-1]=0;
					
				}else
				if(!strcasecmp(tag.name,"robots")&&CurSrv->use_robots&&tag.content){
					char * lt;
					char * s;
					s=UdmGetWord(tag.content,&lt,Indexer->Conf->local_charset);
					while(s){
						if(!strcasecmp(s,"ALL")){
							follow=1;index=1;
						}else
						if(!strcasecmp(s,"NONE")){
							follow=0;index=0;
						}else
						if(!strcasecmp(s,"NOINDEX"))
							index=0;
						else
						if(!strcasecmp(s,"NOFOLLOW"))
							follow=0;
						else
						if(!strcasecmp(s,"INDEX"))
							index=1;
						else
						if(!strcasecmp(s,"FOLLOW"))
							follow=1;
						s=UdmGetWord(NULL,&lt,Indexer->Conf->local_charset);
					}
				}else
				if((!strcasecmp(tag.name,"language"))&&(tag.content)){
					strncpy(Indexer->lang[1].lang,tag.content,2);
					Indexer->lang[1].lang[2]=0;
					Indexer->lang[1].count=1000000; /* This should be enough */
				}
			}
			else	if(!strcmp(s,"img"))	href=tag.src;/*2786*/
			else	if(!strcmp(s,"link"))	href=tag.href;/*2241*/
			else	if(!strcmp(s,"frame"))	href=tag.src;
			else	if(!strcmp(s,"script"))	inscript=opening;
			else	if(!strcmp(s,"style"))	instyle=opening;
			else	if(!strcmp(s,"area"))	href=tag.href;
			else	if(!strcmp(s,"noindex"))comment=opening;
			else	if(!strncasecmp(s,"group_", 6)){
					if(opening)
						if((group = atoi(s+6))>0)
							group=1<<group;
						else
							group = 0;
					else
						group = 0;
				}
			else	if((!strcmp(s,"base"))&&(tag.href)){
				/* <BASE HREF="xxx"> stuff            */
				/* Check that URL is properly formed  */
				/* baseURL is just temporary variable */
				/* If parsing  fails we'll use old    */
				/* base href, passed via CurURL       */
				
				/* Note that we will not check BASE     */
				/* if delete_no_server is unset         */
				/* This is  actually dirty hack. We     */
				/* must check that hostname is the same */

				if((CurSrv->delete_no_server)||(CurSrv->follow&UDM_FOLLOW_WORLD)){
					UDM_URL baseURL;
					int res;
					
					if(!(res=UdmParseURL(&baseURL,tag.href))){
						UdmParseURL(CurURL,tag.href);
					}else{
						switch(res){
						case UDM_PARSEURL_LONG:
							UdmLog(Indexer,UDM_LOG_ERROR,"BASE HREF too long: '%s'",tag.href);
							break;
						case UDM_PARSEURL_BAD:
						default:
							UdmLog(Indexer,UDM_LOG_ERROR,"Error in BASE HREF URL: '%s'",tag.href);
						}
					}
				}
			}

			if((href)&&(follow)&&(CurSrv->follow)){
				UDM_URL newURL;
				int res;
				if(Doc->hops>=CurSrv->maxhops){
					UdmLog(Indexer,UDM_LOG_DEBUG,"Skip \"%s\" : too many hops: %d",href,Doc->hops);
				}else
				if(!(res=UdmParseURL(&newURL,href))){
					char * newschema;
					int Method;
					char reason[UDMSTRSIZ]="";
					if(newURL.schema[0])newschema=newURL.schema;
					else	newschema=CurURL->schema;
					if(!strcasecmp(newschema,"file")||!strcasecmp(newschema,"htdb")){
						sprintf(str,"%s:%s%s",newschema,newURL.path[0]?newURL.path:CurURL->path,newURL.filename);
					}else{
						sprintf(str,"%s://%s%s%s",
						newURL.schema[0]?newURL.schema:CurURL->schema,
						newURL.hostinfo[0]?newURL.hostinfo:CurURL->hostinfo,
						newURL.path[0]?newURL.path:CurURL->path,newURL.filename);
					}
					if((newschema=strchr(str,':')))
						UdmRemove2Dot(newschema+1);
					
					if(!UDM_STRNCMP(str,"ftp://")&&(strstr(str,";type=")))
						*(strstr(str,";type"))=0;

					Method=UdmFindFilter(Indexer->Conf,str,reason);
					UdmLog(Indexer,UDM_LOG_DEBUG,"\"%s\" : '%s'",href,reason);

					if((Method==UDM_DISALLOW)){
					}else
					if((UdmFindRobots(Indexer->Conf,str))&&(CurSrv->use_robots)){
						UdmLog(Indexer,UDM_LOG_DEBUG,"Skip \"%s\" : robots",href);
						UdmLog(Indexer,UDM_LOG_DEBUG,"Full URL \"%s\"",str);
					}else
					if(!UdmFindServer(Indexer->Conf,str)){
						UdmLog(Indexer,UDM_LOG_DEBUG,"Skip \"%s\" : no Server",href);
					}else{
						int add=1;
						/* compare hostinfo in some cases */
						if((!CurSrv->delete_no_server)&&(!(CurSrv->follow&UDM_FOLLOW_WORLD))){
							if(newURL.hostinfo[0])
								add=!strcmp(CurURL->hostinfo,newURL.hostinfo);
						}
						if(add){
						
							if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_LOCK,UDM_LOCK_TARGET);
							
							/* Add URL itself */
							if(UdmAddHref(Indexer->Conf,str,Doc->url_id,Doc->hops+1,0))
								Indexer->Conf->have_targets=1;
							
							/* Add robots.txt for HTTP schema */
							/* When FollowOutside or DeleteNoServer no */
							if((!strcasecmp(newURL.schema,"http"))&&(CurSrv->use_robots)&&
							((CurSrv->follow&UDM_FOLLOW_WORLD)||(!CurSrv->delete_no_server))){
								char str1[UDMSTRSIZ]="";
								sprintf(str1,"%s://%s/%s",newURL.schema,newURL.hostinfo,"robots.txt");
								if(UdmAddHref(Indexer->Conf,str1,0,0,0))
									Indexer->Conf->have_targets=1;
							}
							
							if(Indexer->Conf->LockProc)Indexer->Conf->LockProc(UDM_UNLOCK,UDM_LOCK_TARGET);
						}
					}
				}else{
					switch(res){
					case UDM_PARSEURL_LONG:
						UdmLog(Indexer,UDM_LOG_DEBUG,"URL too long: '%s'",href);
						break;
					case UDM_PARSEURL_BAD:
					default:
						UdmLog(Indexer,UDM_LOG_DEBUG,"Error in URL: '%s'",href);
					}
				}
			}
			UdmFreeTag(&tag);
			free(tmp);
			htok=hend;
			break;
 
		case HTML_COM: /* comment */

			if(!UDM_STRNCASECMP(htok,"<!--UdmComment-->"))
				comment=1;
			else
			if(!UDM_STRNCASECMP(htok,"<!--/UdmComment-->"))
				comment=0;

			while(*htok){
				if(!UDM_STRNCMP(htok,"-->")){
					htok+=3;
					break;
				}
				htok++;
			}
			break;

		case HTML_TXT: /* text */

			if(inscript){
				/* Special case when script  */
				/* body is not commented:    */
				/* <script> x="<"; </script> */
				/* We should find </script>  */
				/* in this case:             */
				for(hend=htok;*hend;hend++){
					if(!UDM_STRNCASECMP(hend,"</script>"))
						break;
				}
			}else{
				for(hend=htok;(*hend)&&(*hend!='<');hend++);
			}
			tmp=(char*)malloc(2*(hend-htok)+6);
			etmp=tmp;*etmp=0;have_space=0;
			for(stmp=htok;stmp<hend;stmp++){
				char sch;
				switch(*stmp){
				case ' ' : case '\n':
				case '\r': case '\t':
					have_space++; break;

				case '&': /* parse specials */
					if(have_space){
						*etmp=' ';etmp++;
						have_space=0;
					}
					stmp++;
					if(*stmp=='#'){ /* &#234; */
						char *code;
						code=++stmp;
						while((stmp<hend)&&(isdigit(*(stmp+1))))
							stmp++;
						if(*(stmp+1)==';')stmp++;
						*etmp=(char)(atoi(code));
						etmp++;*etmp=0;
					}else
					if((sch=UdmSgmlToChar(stmp,Indexer->Conf->local_charset))){
						if(sch==' ')have_space++;
						*etmp=sch;etmp++;*etmp=0;
						while((stmp<hend)&&(isalnum(*(stmp+1))))
							stmp++;
						if(*(stmp+1)==';')stmp++;
					}else{
						*etmp='?';etmp++;*etmp=0;
					}
					break;
				default:
					if(have_space){
						*etmp=' ';etmp++;
						have_space=0;
					}
					*etmp=*stmp;
					etmp++;*etmp=0;
					break;
				}
			}
			if((!comment&&inbody&&!inscript&&!instyle)&&((len=strlen(Doc->text))<UDM_MAXTEXTSIZE-2)){
				len=UDM_MAXTEXTSIZE-2-len;
				if(Doc->text[0])strcat(Doc->text," ");
				strncat(Doc->text,tmp,len);
				Doc->text[UDM_MAXTEXTSIZE-1]=0;
			}
			if((!comment)&&(intitle)&&((len=strlen(Doc->title))<UDM_MAXTITLESIZE-2)){
				len=UDM_MAXTITLESIZE-2-len;
				if(Doc->title[0])strcat(Doc->title," ");
				strncat(Doc->title,tmp,len);
				Doc->title[UDM_MAXTITLESIZE-1]=0;
			}
			if(!intitle&&!comment&&(!hrefonly||group)&&index&&!inscript&&!instyle){
				UdmParseText(Indexer,CurSrv,tmp,group?group:CurSrv->bodyweight*inbody,1);
			}
			free(tmp);
			htok=hend;
			break;
		}
	}
	if(index&&CurSrv->titleweight){
		strcpy(str,Doc->title);
		UdmParseText(Indexer,CurSrv,str,CurSrv->titleweight,1);
	}
	if(index&&CurSrv->keywordweight){
		strcpy(str,Doc->keywords);
		UdmParseText(Indexer,CurSrv,str,CurSrv->keywordweight,1);
	}
	if(index&&CurSrv->descweight){
		strcpy(str,Doc->description);
		UdmParseText(Indexer,CurSrv,str,CurSrv->descweight,1);
	}
	return(0);
}
