#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "udm_config.h"
#include "udm_common.h"
#include "udm_word.h"
#include "udm_db.h"
#include "udm_log.h"
#include "udm_charset.h"
#include "udm_spell.h"


#define RESORT_WORDS	256
#define WSIZE		1024

static int cmpwords(const void * v1, const void * v2){
	return(strcmp(((UDM_WORD*)v1)->word,((UDM_WORD*)v2)->word));
}

static int AddOneWord(UDM_AGENT *Indexer,UDM_SERVER *Server,char *word,int where,int checkstop){
int i;
char *s;
UDM_STOPWORD * stop;
int l,r,c,res,wlen;

	wlen=strlen(word);
	if((wlen>Server->max_word_length)||(wlen<Server->min_word_length))
		return(0);
	s=strdup(word);
	if(strlen(s)>UDM_MAXWORDSIZE)s[UDM_MAXWORDSIZE]=0;
	if(checkstop){
		/* Stopwords checking and language guesser */
		if((stop=UdmIsStopWord(Indexer->Conf,s))){
			int i;
			UdmLog(Indexer,UDM_LOG_DEBUG,"stop: '%s'-'%s'", stop->word, stop->lang);
			/* Add language */
			if(!strcmp(stop->lang,""))
				Indexer->lang[0].count++;
			else
			for(i=1;i<UDM_LANGPERDOC;i++){
				if(!strcmp(Indexer->lang[i].lang,stop->lang)){
					Indexer->lang[i].count++;
					break;
				}
				/* if current lang is free */
				if(!strcmp(Indexer->lang[i].lang,"")){
					strcpy(Indexer->lang[i].lang,stop->lang);
					Indexer->lang[i].count++;
					break;
				}
			}
			free(s);
			return(0);
		}
	}
	/* Find current word in sorted part of word list */
	l=0;r=Indexer->swords-1;
	while(l<=r){
		c=(l+r)/2;
		res=strcmp(Indexer->Word[c].word,s);
		if(res==0){
			Indexer->Word[c].count|=where;
			free(s);
			return(0);
		}
		if(res<0)
			l=c+1;
		else
			r=c-1;
	}
	/* Find in unsorted part */
	for(i=Indexer->swords;i<Indexer->nwords;i++){
		if(!strcmp(Indexer->Word[i].word,s)){
			Indexer->Word[i].count|=where;
			free(s);return(0);
		}
	}
	/* Add new word */
	if(Indexer->nwords>=Indexer->mwords){
		if(Indexer->mwords){
			Indexer->mwords+=WSIZE;
			Indexer->Word=(UDM_WORD *)realloc(Indexer->Word,Indexer->mwords*sizeof(UDM_WORD));
		}else{
			Indexer->mwords=WSIZE;
			Indexer->Word=(UDM_WORD *)malloc(Indexer->mwords*sizeof(UDM_WORD));
		}
	}
	Indexer->Word[Indexer->nwords].word=s;
	Indexer->Word[Indexer->nwords].count=where;
	Indexer->nwords++;

	/* Sort unsorted part */
	if((Indexer->nwords-Indexer->swords)>RESORT_WORDS){
		qsort(Indexer->Word,Indexer->nwords,sizeof(UDM_WORD),cmpwords);
		Indexer->swords=Indexer->nwords;
	}
	return(0);
}

/* This function adds a normalized word form(s) into list using Ispell */
int UdmAddWord(UDM_AGENT *Indexer,UDM_SERVER *Server,char *word,int where,int checkstop){
char 	** forms, ** saveforms;
int	have_digit=0;
int	have_alpha=0;

	if(Server->number_factor==0||Server->alnum_factor==0){
		char *s;
		s=word;
		while(*s){
			if(isdigit(*s))
				have_digit=1;
			else	
				have_alpha=1;
			if(have_digit&&have_alpha)break;
			s++;
		}
		if(have_digit){
			if(have_alpha){
				if(!Server->alnum_factor)return(0);
			}else{
				if(!Server->number_factor)return(0);
			}
		}
	}
	UdmTolower(word,Indexer->Conf->local_charset);
	if((saveforms=forms=UdmNormalizeWord(Indexer,word))){
		/* Add all NORMAL forms of the word */
		while(*forms){
			/* Add only if correct words are allowed */
			if(Server->correct_factor){
				AddOneWord(Indexer,Server,*forms,where,checkstop);
			}
			
			free(*forms);
			forms++;
		}
		free(saveforms);
	}else{
		/* If NORMAL forms has not been found  */
		/*   then we will add the word itself  */
		/* Do it only when incorrect words are */
		/* allowed by configuration            */
		if(Server->incorrect_factor)
			AddOneWord(Indexer,Server,word,where,checkstop);
	}
	return(0);
}

int UdmFreeWords(UDM_AGENT* Indexer) {
	int i;
	for(i=0;i<Indexer->nwords;i++)
		free(Indexer->Word[i].word);
	Indexer->nwords=0;
	Indexer->swords=0;
	return(0);
}
